# ruff: noqa: D100 D101 D103 D105 D107
import numpy as np
from numpy.testing import assert_allclose
from resample import permutation as perm
from scipy import stats
import pytest


@pytest.fixture()
def rng():
    return np.random.Generator(np.random.PCG64(1))


def test_TestResult():
    p = perm.TestResult(1, 2, [3, 4])
    assert p.statistic == 1
    assert p.pvalue == 2
    assert p.samples == [3, 4]
    assert repr(p) == "<TestResult statistic=1 pvalue=2 samples=[3, 4]>"
    assert len(p) == 3
    first, *rest = p
    assert first == 1
    assert rest == [2, [3, 4]]

    p2 = perm.TestResult(1, 2, np.arange(10))
    assert repr(p2) == (
        "<TestResult statistic=1 pvalue=2 samples=[0, 1, 2, ..., 7, 8, 9]>"
    )


class Scipy:
    def __init__(self, **kwargs):
        self.d = kwargs

    def __getitem__(self, key):
        if key in self.d:
            return self.d[key]
        return getattr(stats, key)


scipy = Scipy(
    anova=stats.f_oneway,
    ttest=lambda x, y: stats.ttest_ind(x, y, equal_var=False),
)


@pytest.mark.parametrize(
    "test_name",
    (
        "anova",
        "kruskal",
        "pearsonr",
        "spearmanr",
        "ttest",
    ),
)
@pytest.mark.parametrize("size", (10, 100))
def test_two_sample_same_size(test_name, size, rng):
    x = rng.normal(size=size)
    y = rng.normal(1, size=size)

    test = getattr(perm, test_name)
    scipy_test = scipy[test_name]

    for a, b in ((x, y), (y, x)):
        expected = scipy_test(a, b)
        got = test(a, b, size=999, random_state=1)
        assert_allclose(expected[0], got[0])
        assert_allclose(expected[1], got[1], atol={10: 0.2, 100: 0.02}[size])


@pytest.mark.parametrize(
    "test_name",
    (
        "anova",
        "kruskal",
        "pearsonr",
        "spearmanr",
        "ttest",
    ),
)
@pytest.mark.parametrize("size", (10, 100))
def test_two_sample_different_size(test_name, size, rng):
    x = rng.normal(size=size)
    y = rng.normal(1, size=2 * size)

    test = getattr(perm, test_name)
    scipy_test = scipy[test_name]

    if test_name in ("pearsonr", "spearmanr"):
        with pytest.raises(ValueError):
            test(x, y)
        return

    for a, b in ((x, y), (y, x)):
        expected = scipy_test(a, b)
        got = test(a, b, size=999, random_state=1)
        assert_allclose(expected[0], got[0])
        assert_allclose(expected[1], got[1], atol=5e-2)


@pytest.mark.parametrize(
    "test_name",
    (
        "anova",
        "kruskal",
    ),
)
@pytest.mark.parametrize("size", (10, 100))
def test_three_sample_same_size(test_name, size, rng):
    x = rng.normal(size=size)
    y = rng.normal(1, size=size)
    z = rng.normal(0.5, size=size)

    test = getattr(perm, test_name)
    scipy_test = scipy[test_name]

    for a, b, c in ((x, y, z), (z, y, x)):
        expected = scipy_test(a, b, c)
        got = test(a, b, c, size=999, random_state=1)
        assert_allclose(expected[0], got[0])
        assert_allclose(expected[1], got[1], atol=5e-2)


@pytest.mark.parametrize(
    "test_name",
    (
        "anova",
        "kruskal",
    ),
)
@pytest.mark.parametrize("size", (10, 100))
def test_three_sample_different_size(test_name, size, rng):
    x = rng.normal(size=size)
    y = rng.normal(1, size=2 * size)
    z = rng.normal(0.5, size=size * 2)

    test = getattr(perm, test_name)
    scipy_test = scipy[test_name]

    for a, b, c in ((x, y, z), (z, y, x)):
        expected = scipy_test(a, b, c)
        got = test(a, b, c, size=500, random_state=1)
        assert_allclose(expected[0], got[0])
        assert_allclose(expected[1], got[1], atol=5e-2)


def test_bad_input():
    with pytest.raises(ValueError):
        perm.ttest([1, 2, 3], [1.0, np.nan, 2.0])


@pytest.mark.parametrize("method", ("auto", "patefield", "boyett"))
def test_usp_1(method, rng):
    x = rng.normal(0, 2, size=100)
    y = rng.normal(1, 3, size=100)

    w = np.histogram2d(x, y, bins=(5, 10))[0]
    r = perm.usp(w, method=method, size=100, random_state=1)
    assert r.pvalue > 0.05


@pytest.mark.parametrize("method", ("auto", "patefield", "boyett"))
def test_usp_2(method, rng):
    x = rng.normal(0, 2, size=100).astype(int)

    w = np.histogram2d(x, x, range=((-5, 5), (-5, 5)))[0]

    r = perm.usp(w, method=method, size=99, random_state=1)
    assert r.pvalue == 0.01


@pytest.mark.parametrize("method", ("auto", "patefield", "boyett"))
def test_usp_3(method, rng):
    cov = np.empty((2, 2))
    cov[0, 0] = 2**2
    cov[1, 1] = 3**2
    rho = 0.5
    cov[0, 1] = rho * np.sqrt(cov[0, 0] * cov[1, 1])
    cov[1, 0] = cov[0, 1]

    xy = rng.multivariate_normal([0, 1], cov, size=500).astype(int)

    w = np.histogram2d(*xy.T)[0]

    r = perm.usp(w, method=method, random_state=1)
    assert r.pvalue < 0.0012


@pytest.mark.parametrize("method", ("auto", "patefield", "boyett"))
def test_usp_4(method):
    # table1 from https://doi.org/10.1098/rspa.2021.0549
    w = [[18, 36, 21, 9, 6], [12, 36, 45, 36, 21], [6, 9, 9, 3, 3], [3, 9, 9, 6, 3]]
    r1 = perm.usp(w, method=method, size=9999, random_state=1)
    r2 = perm.usp(np.transpose(w), method=method, size=1, random_state=1)
    assert_allclose(r1.statistic, r2.statistic)
    expected = 0.004106  # checked against USP R package
    assert_allclose(r1.statistic, expected, atol=1e-6)
    # according to paper, pvalue is 0.001, but USP R package gives correct value
    expected = 0.0024  # computed from USP R package with b=99999
    assert_allclose(r1.pvalue, expected, atol=0.001)


@pytest.mark.parametrize("method", ("auto", "patefield", "boyett"))
def test_usp_5(method, rng):
    w = np.empty((100, 100))
    for i in range(100):
        for j in range(100):
            w[i, j] = (i + j) % 2
    r = perm.usp(w, method=method, size=99, random_state=1)
    assert r.pvalue > 0.1


def test_usp_bias(rng):
    # We compute the p-value as an upper limit to the type I error rate.
    # Therefore, the p-value is not unbiased. For size=1, we expect
    # an average p-value = (1 + 0.5) / (1 + 1) = 0.75
    got = [
        perm.usp(rng.poisson(1000, size=(2, 2)), size=1, random_state=i).pvalue
        for i in range(1000)
    ]
    assert_allclose(np.mean(got), 0.75, atol=0.05)


def test_usp_bad_input():
    with pytest.raises(ValueError):
        perm.usp([[1, 2], [3, 4]], size=0)

    with pytest.raises(ValueError):
        perm.usp([[1, 2], [3, 4]], size=-1)

    with pytest.raises(ValueError):
        perm.usp([1, 2])

    with pytest.raises(ValueError):
        perm.usp([[1, 2], [3, 4]], method="foo")


def test_usp_deprecrated():
    w = [[1, 2, 3], [4, 5, 6]]
    r1 = perm.usp(w, method="boyett", size=100, random_state=1)
    with pytest.warns(FutureWarning):
        r2 = perm.usp(w, method="shuffle", size=100, random_state=1)
    assert r1.statistic == r2.statistic


def test_ttest_bad_input():
    with pytest.raises(ValueError):
        perm.ttest([1, 2], [3, 4], size=0)

    with pytest.raises(ValueError):
        perm.ttest([1, 2], [3, 4], size=-1)

    with pytest.raises(ValueError):
        perm.ttest(1, 2)

    with pytest.raises(ValueError):
        perm.ttest([1], [2])



"""
Permutation-based tests.

A collection of statistical tests that use permutated samples. Permutations are used to
compute the distribution of a test statistic under some null hypothesis to obtain
p-values without relying on approximate asymptotic formulas.

The permutation method is generic, it can be used with any test statistic, therefore we
also provide a generic test function that accepts a user-defined function to compute the
test statistic and then automatically computes the p-value for that statistic. The other
tests internally also call this generic test function.

All tests return a TestResult object, which mimics the interface of the result
objects returned by tests in scipy.stats, but has a third field to return the
estimated distribution of the test statistic under the null hypothesis.

Further reading:

- https://en.wikipedia.org/wiki/P-value
- https://en.wikipedia.org/wiki/Test_statistic
- https://en.wikipedia.org/wiki/Paired_difference_test
"""

__all__ = [
    "TestResult",
    "usp",
    "same_population",
    "anova",
    "kruskal",
    "pearsonr",
    "spearmanr",
    "ttest",
]

import sys
import warnings
from dataclasses import dataclass
from typing import Any, Callable, Optional, Tuple, Union

import numpy as np
from numpy.typing import ArrayLike, NDArray
from scipy import stats as _stats

from . import _util

_dataclass_kwargs = {"frozen": True, "repr": False}
if sys.version_info >= (3, 10):
    _dataclass_kwargs["slots"] = True  # pragma: no cover


@dataclass(**_dataclass_kwargs)
class TestResult:
    """
    Holder of the result of the permutation test.

    This class acts like a tuple, which means its can be unpacked and the fields can be
    accessed by name or by index.

    Attributes
    ----------
    statistic: float
        Value of the test statistic computed on the original data
    pvalue: float
        Estimated chance probability (aka Type I error) for rejecting the null
        hypothesis. See https://en.wikipedia.org/wiki/P-value for details.
    samples: array
        Values of the test statistic from the permutated samples.

    """

    statistic: float
    pvalue: float
    samples: NDArray

    def __repr__(self) -> str:
        """Return (potentially shortened) representation."""
        s = None
        if len(self.samples) < 7:
            s = str(self.samples)
        else:
            s = "[{}, {}, {}, ..., {}, {}, {}]".format(
                *self.samples[:3], *self.samples[-3:]
            )
        return (
            f"<TestResult statistic={self.statistic} pvalue={self.pvalue} samples={s}>"
        )

    def __len__(self) -> int:
        """Return length of tuple."""
        return 3

    def __getitem__(self, idx: int) -> Union[float, NDArray]:
        """Return fields by index."""
        if idx == 0:
            return self.statistic
        elif idx == 1:
            return self.pvalue
        elif idx == 2:
            return self.samples
        raise IndexError


def usp(
    w: "ArrayLike",
    *,
    size: int = 9999,
    method: str = "auto",
    random_state: Optional[Union[np.random.Generator, int]] = None,
) -> TestResult:
    """
    Test independence of two discrete data sets with the U-statistic.

    The USP test is described in this paper: https://doi.org/10.1098/rspa.2021.0549.
    According to the paper, it outperforms the Pearson's χ² and the G-test in both in
    stability and power.

    It requires that the input is a contigency table (a 2D histogram of value pairs).
    Whether the original values were discrete or continuous does not matter for the
    test. In case of continuous values, using a large number of bins is safe, since the
    test is not negatively affected by bins with zero entries.

    Parameters
    ----------
    w : array-like
        Two-dimensional array which represents the counts in a histogram. The counts
        can be of floating point type, but must have integral values.
    size : int, optional
        Number of permutations. Default 9999.
    method : str, optional
        Method used to generate random tables under the null hypothesis.
        'auto': Use heuristic to select fastest algorithm for given table.
        'boyett': Boyett's algorithm, which requires extra space to store N + 1
        integers for N entries in total and has O(N) time complexity. It performs
        poorly when N is large, but does not depend on the number of K table cells.
        'patefield': Patefield's algorithm, which does not require extra space and
        has O(K log(N)) time complexity. It performs well even if N is huge. For
        small N and large K, the shuffling algorithm is faster.
        Default is 'auto'.
    random_state : numpy.random.Generator or int, optional
        Random number generator instance. If an integer is passed, seed the numpy
        default generator with it. Default is to use ``numpy.random.default_rng()``.

    Returns
    -------
    TestResult

    """
    if size <= 0:
        raise ValueError("size must be positive")

    if method == "shuffle":
        warnings.warn(
            "method 'shuffle' is deprecated, please use 'boyett'", FutureWarning
        )
        method = "boyett"

    rng = _util.normalize_rng(random_state)

    w = np.array(w, dtype=float)
    if w.ndim != 2:
        raise ValueError("w must be two-dimensional")

    r = np.sum(w, axis=1)
    c = np.sum(w, axis=0)
    ntot = np.sum(r)

    m = np.outer(r, c) / ntot

    f1 = 1.0 / (ntot * (ntot - 3))
    f2 = 4.0 / (ntot * (ntot - 2) * (ntot - 3))

    t = _usp(f1, f2, w, m)

    ts = np.empty(size)
    for b, w in enumerate(
        _stats.random_table(r, c).rvs(
            size, method=None if method == "auto" else method, random_state=rng
        )
    ):
        # m stays the same, since r and c remain unchanged
        ts[b] = _usp(f1, f2, w, m)

    # Thomas B. Berrett, Ioannis Kontoyiannis, Richard J. Samworth
    # Ann. Statist. 49(5): 2457-2490 (October 2021). DOI: 10.1214/20-AOS2041
    # Eq. 5 says we need to add 1 to n_pass and n_total
    pvalue = (np.sum(t <= ts) + 1) / (size + 1)

    return TestResult(t, pvalue, ts)


def _usp(f1: float, f2: float, w: NDArray, m: NDArray) -> NDArray:
    # Eq. 2.1 from https://doi.org/10.1098/rspa.2021.0549
    return f1 * np.sum((w - m) ** 2) - f2 * np.sum(w * m)


def same_population(
    fn: Callable[..., float],
    x: "ArrayLike",
    y: "ArrayLike",
    *args: "ArrayLike",
    transform: Optional[Callable[[NDArray], NDArray]] = None,
    size: int = 9999,
    random_state: Optional[Union[np.random.Generator, int]] = None,
) -> TestResult:
    """
    Compute p-value for hypothesis that samples originate from same population.

    The computation is based on a user-defined test statistic. The distribution of the
    test statistic under the null hypothesis is generated by generating random
    permutations of the inputs, to simulate that they are actually drawn from the same
    population. The test statistic is recomputed on these permutations and the p-value
    is computed as the fraction of these resampled test statistics which are larger
    than the original value.

    Some test statistics need to be transformed to fulfill the condition above, for
    example if they are signed. A transform can be passed to this function for those
    cases.

    Parameters
    ----------
    fn : Callable
        Function with signature f(x, ...), where the number of arguments corresponds to
        the number of data samples passed to the test.
    x : array-like
        First sample.
    y : array-like
        Second sample.
    *args: array-like
        Further samples, if the test allows to compare more than two.
    transform : Callable, optional
        Function with signature f(x) for the test statistic to turn it into a measure of
        deviation. Must be vectorised.
    size : int, optional
        Number of permutations. Default 9999.
    random_state : numpy.random.Generator or int, optional
        Random number generator instance. If an integer is passed, seed the numpy
        default generator with it. Default is to use `numpy.random.default_rng()`.

    Returns
    -------
    TestResult

    """
    if size <= 0:
        raise ValueError("max_size must be positive")

    rng = _util.normalize_rng(random_state)

    r = []
    for arg in (x, y) + args:
        a = np.array(arg)
        if a.ndim != 1:
            raise ValueError("input samples must be 1D arrays")
        if len(a) < 2:
            raise ValueError("input arrays must have at least two items")
        if a.dtype.kind == "f" and np.any(np.isnan(a)):
            raise ValueError("input contains NaN")
        r.append(a)
    args = r
    del r

    # compute test statistic for original input
    t = fn(*args)

    # compute test statistic for permutated inputs
    slices = []
    start = 0
    for a in args:
        stop = start + len(a)
        slices.append(slice(start, stop))
        start = stop

    joined_sample = np.concatenate(args)

    # For algorithm below, see comment in usp function.
    ts = np.empty(size)
    for b in range(size):
        rng.shuffle(joined_sample)
        ts[b] = fn(*(joined_sample[sl] for sl in slices))
    if transform is None:
        u = t
        us = ts
    else:
        u = transform(t)
        us = transform(ts)
    # see usp for why we need to add 1 to both numerator and denominator
    pvalue = (np.sum(u <= us) + 1) / (size + 1)

    return TestResult(t, pvalue, ts)


def anova(
    x: "ArrayLike", y: "ArrayLike", *args: "ArrayLike", **kwargs: Any
) -> TestResult:
    """
    Test whether the means of two or more samples are compatible.

    This test uses one-way analysis of variance (one-way ANOVA) which tests whether the
    samples have the same mean. This test is typically used when one has three groups
    or more. For two groups, Welch's ttest is preferred, because ANOVA assumes equal
    variances for the samples.

    Parameters
    ----------
    x : array-like
        First sample.
    y : array-like
        Second sample.
    *args : array-like
        Further samples.
    **kwargs :
        Keyword arguments are forward to :meth:`same_population`.

    Returns
    -------
    TestResult

    Notes
    -----
    https://en.wikipedia.org/wiki/One-way_analysis_of_variance
    https://en.wikipedia.org/wiki/F-test

    """
    kwargs["transform"] = None
    return same_population(_ANOVA(), x, y, *args, **kwargs)


def kruskal(
    x: "ArrayLike", y: "ArrayLike", *args: "ArrayLike", **kwargs: Any
) -> TestResult:
    """
    Test whether two or more samples have the same mean rank.

    This performs a permutation-based Kruskal-Wallis test. In a sense, it extends the
    Mann-Whitney U test, which also uses ranks, to more than two groups. It does so by
    comparing the means of the rank distributions.

    Parameters
    ----------
    x : array-like
        First sample.
    y : array-like
        Second sample.
    *args : array-like
        Further samples.
    **kwargs :
        Keyword arguments are forward to :meth:`same_population`.

    Returns
    -------
    TestResult

    Notes
    -----
    https://en.wikipedia.org/wiki/Kruskal%E2%80%93Wallis_one-way_analysis_of_variance

    """
    kwargs["transform"] = None
    return same_population(_kruskal, x, y, *args, **kwargs)


def pearsonr(x: "ArrayLike", y: "ArrayLike", **kwargs: Any) -> TestResult:
    """
    Test whether two samples are drawn from same population using correlation.

    The test statistic is the Pearson correlation coefficient. The test is very
    sensitive to linear relationship of x and y. If the relationship is very non-linear
    but monotonic, :func:`spearmanr` may be more sensitive.

    https://en.wikipedia.org/wiki/Pearson_correlation_coefficient

    Parameters
    ----------
    x : array-like
        First sample.
    y : array-like
        Second sample.
    **kwargs :
        Keyword arguments are forward to :meth:`same_population`.

    Returns
    -------
    TestResult

    """
    if len(x) != len(y):
        raise ValueError("x and y must have have the same length")
    kwargs["transform"] = np.abs
    return same_population(_pearson, x, y, **kwargs)


def spearmanr(x: "ArrayLike", y: "ArrayLike", **kwargs: Any) -> TestResult:
    """
    Test whether two samples are drawn from same population using rank correlation.

    The test statistic is Spearman's rank correlation coefficient. The test is very
    sensitive to monotonic relationships between x and y, even if it is very non-linear.

    Parameters
    ----------
    x : array-like
        First sample.
    y : array-like
        Second sample.
    **kwargs :
        Keyword arguments are forward to :meth:`same_population`.

    Returns
    -------
    TestResult

    """
    if len(x) != len(y):
        raise ValueError("x and y must have have the same length")
    kwargs["transform"] = np.abs
    return same_population(_spearman, x, y, **kwargs)


def ttest(x: "ArrayLike", y: "ArrayLike", **kwargs: Any) -> TestResult:
    """
    Test whether the means of two samples are compatible with Welch's t-test.

    See https://en.wikipedia.org/wiki/Welch%27s_t-test for details on this test. The
    p-value computed is for the null hypothesis that the two population means are equal.
    The test is two-sided, which means that swapping x and y gives the same p-value.
    Welch's t-test does not require the sample sizes to be equal and it does not require
    the samples to have the same variance.

    Parameters
    ----------
    x : array-like
        First sample.
    y : array-like
        Second sample.
    **kwargs :
        Keyword arguments are forward to :meth:`same_population`.

    Returns
    -------
    TestResult

    """
    kwargs["transform"] = np.abs
    return same_population(_ttest, x, y, **kwargs)


def _ttest(x: NDArray, y: NDArray) -> float:
    n1 = len(x)
    n2 = len(y)
    m1 = np.mean(x)
    m2 = np.mean(y)
    v1 = np.var(x, ddof=1)
    v2 = np.var(y, ddof=1)
    r: float = (m1 - m2) / np.sqrt(v1 / n1 + v2 / n2)
    return r


def _pearson(x: NDArray, y: NDArray) -> float:
    m1 = np.mean(x)
    m2 = np.mean(y)
    s1 = np.mean((x - m1) ** 2)
    s2 = np.mean((y - m2) ** 2)
    r: float = np.mean((x - m1) * (y - m2)) / np.sqrt(s1 * s2)
    return r


def _spearman(x: NDArray, y: NDArray) -> float:
    x = _stats.rankdata(x)
    y = _stats.rankdata(y)
    return _pearson(x, y)


def _kruskal(*args: NDArray) -> float:
    # see https://en.wikipedia.org/wiki/
    #           Kruskal%E2%80%93Wallis_one-way_analysis_of_variance
    # method 3 and 4
    joined = np.concatenate(args)
    r = _stats.rankdata(joined)
    n = len(r)
    start = 0
    r_args = []
    for i, a in enumerate(args):
        r_args.append(r[start : start + len(a)])
        start += len(a)

    # method 3 (assuming no ties)
    h: float = 12.0 / (n * (n + 1)) * sum(
        len(r) * np.mean(r) ** 2 for r in r_args
    ) - 3 * (n + 1)

    # apply tie correction
    h /= _stats.tiecorrect(r)
    return h


class _ANOVA:
    # see https://en.wikipedia.org/wiki/F-test
    km1: int = -2
    nmk: int = 0
    a_bar: float = 0.0

    def __call__(self, *args: NDArray) -> float:
        if self.km1 == -2:
            self._init(args)

        between_group_variability: float = (
            sum(len(a) * (np.mean(a) - self.a_bar) ** 2 for a in args) / self.km1
        )
        within_group_variability: float = sum(len(a) * np.var(a) for a in args) / (
            self.nmk
        )
        return between_group_variability / within_group_variability

    def _init(self, args: Tuple[NDArray, ...]) -> None:
        n = sum(len(a) for a in args)
        k = len(args)
        self.km1 = k - 1
        self.nmk = n - k
        self.a_bar = np.mean(np.concatenate(args))



"""
Jackknife resampling tools.

Compute estimator bias and variance with jackknife resampling. The implementation
supports resampling of N-dimensional data. The interface of this module mimics that of
the bootstrap module, so that you can easily switch between bootstrapping and
jackknifing bias and variance of an estimator.

The jackknife is an approximation to the bootstrap, so in general bootstrapping is
preferred, especially when the sample is small. The computational cost of the jackknife
increases quadratically with the sample size, but only linearly for the bootstrap. An
advantage of the jackknife can be the deterministic outcome, since no random sampling
is involved, but this can be overcome by fixing the seed for the bootstrap.

The jackknife should be used to estimate the bias, since the bootstrap cannot (easily)
estimate bias. The bootstrap should be preferred when computing the variance.
"""

__all__ = [
    "resample",
    "jackknife",
    "bias",
    "bias_corrected",
    "variance",
    "cross_validation",
]

from typing import Any, Callable, Collection, Generator, List

import numpy as np
from numpy.typing import ArrayLike


def resample(
    sample: "ArrayLike", *args: "ArrayLike", copy: bool = True
) -> Generator[Any, None, None]:
    """
    Generate jackknifed samples.

    Parameters
    ----------
    sample : array-like
        Sample. If the sequence is multi-dimensional, the first dimension must
        walk over i.i.d. observations.
    *args: array-like
        Optional additional arrays of the same length to resample.
    copy : bool, optional
        If `True`, return the replicated sample as a copy, otherwise return a view into
        the internal array buffer of the generator. Setting this to `False` avoids
        `len(sample)` copies, which is more efficient, but see notes for caveats.

    Yields
    ------
    ndarray
        Array with same shape and type as input, but with the size of the first
        dimension reduced by one. Replicates are missing one value of the original in
        ascending order, e.g. for a sample (1, 2, 3), one gets (2, 3), (1, 3), (1, 2).

    See Also
    --------
    resample.bootstrap.resample : Generate bootstrap samples.
    resample.jackknife.jackknife : Generate jackknife estimates.

    Notes
    -----
    On performance:

    The generator interally keeps a single array to the replicates, which is updated
    on each iteration of the generator. The safe default is to return copies of this
    internal state. To increase performance, it also possible to return a view into
    the generator state, by setting the `copy=False`. However, this will only produce
    correct results if the generator is called strictly sequentially in a single-
    threaded program and the loop body consumes the view and does not try to store it.
    The following program shows what happens otherwise:

    >>> from resample.jackknife import resample
    >>> r1 = []
    >>> for x in resample((1, 2, 3)): # works as expected
    ...     r1.append(x)
    >>> print(r1)
    [array([2, 3]), array([1, 3]), array([1, 2])]
    >>>
    >>> r2 = []
    >>> for x in resample((1, 2, 3), copy=False):
    ...     r2.append(x) # x is now a view into the same array in memory
    >>> print(r2)
    [array([1, 2]), array([1, 2]), array([1, 2])]

    """
    sample_np = np.atleast_1d(sample)
    n_sample = len(sample_np)

    args_np = []
    if args:
        if not isinstance(args[0], Collection):
            import warnings

            warnings.warn(
                "Calling resample with positional instead of keyword arguments is "
                "deprecated",
                FutureWarning,
            )
            if len(args) == 1:
                (copy,) = args
            else:
                raise ValueError("too many arguments")
        else:
            args_np.append(sample_np)
            for arg in args:
                arg_np = np.atleast_1d(arg)
                n_arg = len(arg_np)
                if n_arg != n_sample:
                    raise ValueError(
                        f"extra argument has wrong length {n_arg} != {n_sample}"
                    )
                args_np.append(arg_np)

    if args_np:
        return _resample_n(args_np, copy)
    return _resample_1(sample_np, copy)


def _resample_1(sample: np.ndarray, copy: bool) -> Generator[np.ndarray, None, None]:
    # yield x0
    x = sample[1:].copy()
    yield x.copy() if copy else x

    # update of x needs to change only value at index i
    # for a = [0, 1, 2, 3]
    # x0 = [1, 2, 3] (yielded above)
    # x1 = [0, 2, 3] # override first index
    # x2 = [0, 1, 3] # override second index
    # x3 = [0, 1, 2] # ...
    for i in range(len(sample) - 1):
        x[i] = sample[i]
        yield x.copy() if copy else x


def _resample_n(samples: List[np.ndarray], copy: bool) -> Generator[Any, None, None]:
    x = [a[1:].copy() for a in samples]
    yield (xi.copy() for xi in x)
    for i in range(len(samples[0]) - 1):
        for xi, ai in zip(x, samples):
            xi[i] = ai[i]
        yield (xi.copy() for xi in x)


def jackknife(
    fn: Callable[..., np.ndarray],
    sample: "ArrayLike",
    *args: "ArrayLike",
) -> np.ndarray:
    """
    Calculate jackknife estimates for a given sample and estimator.

    The jackknife is a linear approximation to the bootstrap. In contrast to the
    bootstrap it is deterministic and does not use random numbers. The caveat is the
    computational cost of the jackknife, which is O(n²) for n observations, compared
    to O(n x k) for k bootstrap replicates. For large samples, the bootstrap is more
    efficient.

    Parameters
    ----------
    fn : callable
        Estimator. Can be any mapping ℝⁿ → ℝᵏ, where n is the sample size
        and k is the length of the output array.
    sample : array-like
        Original sample.
    *args: array-like
        Optional additional arrays of the same length to resample.

    Returns
    -------
    ndarray
        Jackknife samples.

    Examples
    --------
    >>> from resample.jackknife import jackknife
    >>> import numpy as np
    >>> x = np.arange(10)
    >>> fx = np.mean(x)
    >>> fb = jackknife(np.mean, x)
    >>> print(f"f(x) = {fx:.1f} +/- {np.std(fb):.1f}")
    f(x) = 4.5 +/- 0.3

    """
    gen = resample(sample, *args, copy=False)
    if args:
        return np.array([fn(*b) for b in gen])
    return np.asarray([fn(b) for b in gen])


def bias(
    fn: Callable[..., np.ndarray], sample: "ArrayLike", *args: "ArrayLike"
) -> np.ndarray:
    """
    Calculate jackknife estimate of bias.

    The bias estimate is accurate to O(1/n), where n is the number of samples.
    If the bias is exactly O(1/n), then the estimate is exact.

    Wikipedia:
    https://en.wikipedia.org/wiki/Jackknife_resampling

    Parameters
    ----------
    fn : callable
        Estimator. Can be any mapping ℝⁿ → ℝᵏ, where n is the sample size
        and k is the length of the output array.
    sample : array-like
        Original sample.
    *args: array-like
        Optional additional arrays of the same length to resample.

    Returns
    -------
    ndarray
        Jackknife estimate of bias (= expectation of estimator - true value).

    Examples
    --------
    Compute bias of numpy.var with and without bias-correction.

    >>> from resample.jackknife import bias
    >>> import numpy as np
    >>> x = np.arange(10)
    >>> b1 = bias(np.var, x)
    >>> b2 = bias(lambda x: np.var(x, ddof=1), x)
    >>> f"bias of naive sample variance {b1:.1f}, bias of corrected variance {b2:.1f}"
    'bias of naive sample variance -0.9, bias of corrected variance 0.0'

    """
    sample = np.atleast_1d(sample)
    n = len(sample)
    theta = fn(sample)
    mean_theta = np.mean(jackknife(fn, sample, *args), axis=0)
    return (n - 1) * (mean_theta - theta)


def bias_corrected(
    fn: Callable[..., np.ndarray], sample: "ArrayLike", *args: "ArrayLike"
) -> np.ndarray:
    """
    Calculate bias-corrected estimate of the function with the jackknife.

    Removes a bias of O(1/n), where n is the sample size, leaving bias of
    order O(1/n²). If the original function has a bias of exactly O(1/n),
    the corrected result is now unbiased.

    Wikipedia:
    https://en.wikipedia.org/wiki/Jackknife_resampling

    Parameters
    ----------
    fn : callable
        Estimator. Can be any mapping ℝⁿ → ℝᵏ, where n is the sample size
        and k is the length of the output array.
    sample : array-like
        Original sample.
    *args: array-like
        Optional additional arrays of the same length to resample.

    Returns
    -------
    ndarray
        Estimate with O(1/n) bias removed.

    Examples
    --------
    Compute bias-corrected estimate of numpy.var.

    >>> from resample.jackknife import bias_corrected
    >>> import numpy as np
    >>> x = np.arange(10)
    >>> v1 = np.var(x)
    >>> v2 = bias_corrected(np.var, x)
    >>> f"naive variance {v1:.1f}, bias-corrected variance {v2:.1f}"
    'naive variance 8.2, bias-corrected variance 9.2'

    """
    sample = np.atleast_1d(sample)
    n = len(sample)
    theta = fn(sample)
    mean_theta = np.mean(jackknife(fn, sample, *args), axis=0)
    return n * theta - (n - 1) * mean_theta


def variance(
    fn: Callable[..., np.ndarray], sample: "ArrayLike", *args: "ArrayLike"
) -> np.ndarray:
    """
    Calculate jackknife estimate of variance.

    Wikipedia:
    https://en.wikipedia.org/wiki/Jackknife_resampling

    Parameters
    ----------
    fn : callable
        Estimator. Can be any mapping ℝⁿ → ℝᵏ, where n is the sample size
        and k is the length of the output array.
    sample : array-like
        Original sample.
    *args: array-like
        Optional additional arrays of the same length to resample.

    Returns
    -------
    ndarray
        Jackknife estimate of variance.

    Examples
    --------
    Compute variance of arithmetic mean.

    >>> from resample.jackknife import variance
    >>> import numpy as np
    >>> x = np.arange(10)
    >>> v = variance(np.mean, x)
    >>> f"{v:.1f}"
    '0.9'

    """
    # formula is (n - 1) / n * sum((fj - mean(fj)) ** 2)
    #   = np.var(fj, ddof=0) * (n - 1)
    sample = np.atleast_1d(sample)
    thetas = jackknife(fn, sample, *args)
    n = len(sample)
    return (n - 1) * np.var(thetas, ddof=0, axis=0)


def cross_validation(
    predict: Callable[..., float], x: "ArrayLike", y: "ArrayLike", *args: "ArrayLike"
) -> float:
    """
    Calculate mean-squared error of model with leave-one-out-cross-validation.

    Wikipedia:
    https://en.wikipedia.org/wiki/Cross-validation_(statistics)

    Parameters
    ----------
    predict : callable
        Function with the signature (x_in, y_in, x_out, *args). It takes x_in, y_in,
        which are arrays with the same length. x_out should be one element of the x
        array. *args are further optional arguments for the function. The function
        should return the prediction y(x_out).
    x : array-like
        Explanatory variable. Must be an array of shape (N, ...), where N is the number
        of samples.
    y : array-like
        Observations. Must be an array of shape (N, ...).
    *args:
        Optional arguments which are passed unmodified to predict.

    Returns
    -------
    float
        Variance of the difference (y[i] - predict(..., x[i], *args)).

    """
    deltas = []
    for i, (x_in, y_in) in enumerate(resample(x, y, copy=False)):
        yip = predict(x_in, y_in, x[i], *args)
        deltas.append((y[i] - yip))
    return np.var(deltas)  # type:ignore



# ruff: noqa: D100 D103
import numpy as np
import pytest
from numpy.testing import assert_equal
from resample.jackknife import resample


def run_resample(n, copy):
    x = np.arange(n)
    r = []
    for b in resample(x, copy=copy):
        r.append(np.mean(b))
    return r


@pytest.mark.benchmark(group="jackknife-100")
@pytest.mark.parametrize("copy", (True, False))
def test_jackknife_resample_100(benchmark, copy):
    result = benchmark(run_resample, 100, copy)
    assert_equal(result, run_resample(100, resample))


@pytest.mark.benchmark(group="jackknife-1000")
@pytest.mark.parametrize("copy", (True, False))
def test_jackknife_resample_1000(benchmark, copy):
    result = benchmark(run_resample, 1000, copy)
    assert_equal(result, run_resample(1000, resample))



import matplotlib.pyplot as plt
import json
from pathlib import Path
import numpy as np

d = Path(__file__).parent if "__file__" in globals() else Path()

with open(d / "bench_rcont.json") as f:
    data = json.load(f)

vs = [[[], [], []], [[], [], []]]

benchs = data["benchmarks"]
for b in benchs:
    params = b["params"]
    m = params["method"]
    n = params["n"]
    k = params["k"]
    stats = b["stats"]
    val = stats["mean"]
    err = stats["stddev"] / stats["rounds"] ** 0.5
    vs[m][0].append(n)
    vs[m][1].append(k)
    vs[m][2].append(val)

fig, ax = plt.subplots(1, 2, figsize=(10, 4), sharey=True)
for i, label in enumerate(("shuffle", "patefield")):
    ax[0].scatter(vs[i][0], vs[i][2], s=np.add(vs[i][1], 1), label=label)
    ax[1].scatter(vs[i][1], vs[i][2], s=10 * np.log(vs[i][0]) - 10)
ax[0].loglog()
ax[1].loglog()
ax[0].set_xlabel("N")
ax[1].set_xlabel("K")
ax[0].set_ylabel("t/sec")
plt.figlegend(loc="upper center", ncol=2, frameon=False)



from typing import Optional, Tuple, Union

import numpy as np
from numpy.typing import ArrayLike

__all__ = ["normalize_rng", "wilson_score_interval"]


def normalize_rng(
    random_state: Optional[Union[int, np.random.Generator]],
) -> np.random.Generator:
    """Return normalized RNG object."""
    if random_state is None:
        return np.random.default_rng()
    if isinstance(random_state, np.random.Generator):
        return random_state
    return np.random.default_rng(random_state)


def wilson_score_interval(
    n1: "ArrayLike", n: "ArrayLike", z: float
) -> Tuple[np.ndarray, Tuple[np.ndarray, np.ndarray]]:
    """Return binomial fraction and Wilson score interval."""
    p = n1 / n
    norm = 1 / (1 + z**2 / n)
    a = p + 0.5 * z**2 / n
    b = z * np.sqrt(p * (1 - p) / n + 0.25 * (z / n) ** 2)
    return p, ((a - b) * norm, (a + b) * norm)



"""
Bootstrap resampling tools.

Compute estimator bias, variance, confidence intervals with bootstrap resampling.

Several forms of bootstrapping on N-dimensional data are supported: ordinary, balanced,
extended, parametric, and stratified sampling, see :func:`resample` for details.
Parametric bootstrapping fits a user-specified distribution to the data and samples
from the parametric distribution. The distributions are taken from scipy.stats.

Confidence intervals can be computed with the ordinary percentile method and with the
more efficient BCa method, see :func:`confidence_interval` for details.
"""

__all__ = [
    "resample",
    "bootstrap",
    "variance",
    "covariance",
    "confidence_interval",
]

from typing import (
    Any,
    Callable,
    Collection,
    Dict,
    Generator,
    List,
    Optional,
    Tuple,
    Union,
)

import numpy as np
from numpy.typing import ArrayLike
from scipy import stats

from . import _util
from .empirical import quantile_function_gen
from .jackknife import jackknife


def resample(
    sample: "ArrayLike",
    *args: "ArrayLike",
    size: int = 100,
    method: str = "balanced",
    strata: Optional["ArrayLike"] = None,
    random_state: Optional[Union[np.random.Generator, int]] = None,
) -> Generator[np.ndarray, None, None]:
    """
    Return generator of bootstrap samples.

    Parameters
    ----------
    sample : array-like
        Original sample.
    *args : array-like
        Optional additional arrays of the same length to resample.
    size : int, optional
        Number of bootstrap samples to generate. Default is 100.
    method : str or None, optional
        How to generate bootstrap samples. Supported are 'ordinary', 'balanced',
        'extended', or a distribution name for a parametric bootstrap.
        Default is 'balanced'. Supported distribution names: 'normal' (also:
        'gaussian', 'norm'), 'student' (also: 't'), 'laplace', 'logistic',
        'F' (also: 'f'), 'beta', 'gamma', 'log-normal' (also: 'lognorm',
        'log-gaussian'), 'inverse-gaussian' (also: 'invgauss'), 'pareto', 'poisson'.
    strata : array-like, optional
        Stratification labels. Must have the same shape as `sample`. Default is None.
    random_state : numpy.random.Generator or int, optional
        Random number generator instance. If an integer is passed, seed the numpy
        default generator with it. Default is to use `numpy.random.default_rng()`.

    Yields
    ------
    ndarray
        Bootstrap sample.

    Examples
    --------
    Compute uncertainty of arithmetic mean.

    >>> from resample.bootstrap import resample
    >>> import numpy as np
    >>> x = np.arange(10)
    >>> fx = np.mean(x)
    >>> fb = []
    >>> for b in resample(x, size=10000, random_state=1):
    ...     fb.append(np.mean(b))
    >>> print(f"f(x) = {fx:.1f} +/- {np.std(fb):.1f}")
    f(x) = 4.5 +/- 0.9

    Compute uncertainty of function applied to multivariate data.

    >>> from resample.bootstrap import resample
    >>> import numpy as np
    >>> x = np.arange(10)
    >>> y = np.arange(10, 20)
    >>> fx = np.mean((x, y))
    >>> fb = []
    >>> for bx, by in resample(x, y, size=10000, random_state=1):
    ...     fb.append(np.mean((bx, by)))
    >>> print(f"f(x, y) = {fx:.1f} +/- {np.std(fb):.1f}")
    f(x, y) = 9.5 +/- 0.9

    Notes
    -----
    Balanced vs. ordinary bootstrap:

    The balanced bootstrap produces more accurate results for the same number of
    bootstrap samples than the ordinary bootstrap, but needs to allocate memory for
    B integers, where B is the number of bootstrap samples. Since values of B larger
    than 10000 are rarely needed, this is usually not an issue.

    Non-parametric vs. parametric bootstrap:

    If you know that the data follow a particular parametric distribution, it is
    better to sample from this parametric distribution, but in most cases it is
    sufficient and more convenient to do a non-parametric bootstrap (using "balanced",
    "ordinary", "extended"). The parametric bootstrap is essential for estimators
    sensitive to the tails of a distribution (for example, a quantile close to 0 or 1).
    In this case, only a parametric bootstrap will give reasonable answers, since the
    non-parametric bootstrap cannot include rare events in the tail if the original
    sample did not have them.

    Extended bootstrap:

    In particle physics and perhaps also in other fields, estimators are used which are
    that are a function of both the size and shape of a sample (for example, fit of a
    peak over smooth background to the mass distribution of decay candidates). In this
    case, the normal bootstrap (parametric or non-parametric) is not correct, since the
    sample size is kept constant. For such cases, one needs the "extended" bootstrap.
    The name alludes to the so-called extended maximum-likelihood (EML) method in
    particle physics. Estimates obtained with the EML need to be bootstrapped with the
    "extended" bootstrap.

    Stratification:

    If the sample consists of several distinct classes, stratification
    ensures that the relative proportions of each class are maintained in each
    replicated sample. This is a stricter constraint than that offered by the
    balanced bootstrap, which only guarantees that classes have the original
    proportions over all replicates, but not within each one replicate.

    """
    sample_np = np.atleast_1d(sample)
    n_sample = len(sample_np)
    args_np: List[np.ndarray] = []

    if args:
        if not isinstance(args[0], Collection):
            import warnings

            warnings.warn(
                "Calling resample with positional instead of keyword parameters is "
                "deprecated",
                FutureWarning,
            )
            kwargs: Dict[str, Any] = {
                "size": size,
                "method": method,
                "strata": strata,
                "random_state": random_state,
            }
            if len(args) > len(kwargs):
                raise ValueError("too many arguments")
            for key, val in zip(kwargs, args):
                kwargs[key] = val
            size = kwargs["size"]
            method = kwargs["method"]
            strata = kwargs["strata"]
            random_state = kwargs["random_state"]
            del args
        else:
            args_np.append(sample_np)
            for arg in args:
                arg = np.atleast_1d(arg)
                n_arg = len(arg)
                if n_arg != n_sample:
                    raise ValueError(
                        f"extra argument has wrong length {n_arg} != {n_sample}"
                    )
                args_np.append(arg)

    rng = _util.normalize_rng(random_state)

    if strata is not None:
        strata_np = np.atleast_1d(strata)
        if args_np:
            raise ValueError("Stratified resampling only works with one sample array")
        if len(strata_np) != n_sample:
            raise ValueError("a and strata must have the same shape")
        return _resample_stratified(sample_np, size, method, strata_np, rng)

    if method == "balanced":
        if args_np:
            return _resample_balanced_n(args_np, size, rng)
        else:
            return _resample_balanced_1(sample_np, size, rng)
    if method == "ordinary":
        if args_np:
            return _resample_ordinary_n(args_np, size, rng)
        else:
            return _resample_ordinary_1(sample_np, size, rng)
    if method == "extended":
        if args_np:
            return _resample_extended_n(args_np, size, rng)
        else:
            return _resample_extended_1(sample_np, size, rng)

    if args_np:
        raise ValueError("Parametric resampling only works with one sample array")

    dist = {
        # put aliases here
        "gaussian": stats.norm,
        "normal": stats.norm,
        "log-normal": stats.lognorm,
        "log-gaussian": stats.lognorm,
        "inverse-gaussian": stats.invgauss,
        "student": stats.t,
    }.get(method)

    # fallback to scipy.stats name
    if dist is None:
        try:
            dist = getattr(stats, method.lower())
        except AttributeError:
            raise ValueError(f"Invalid family: '{method}'")

    if sample_np.ndim > 1:
        if dist != stats.norm:
            raise ValueError(f"family '{method}' only supports 1D samples")
        dist = stats.multivariate_normal
        if sample_np.ndim > 2:
            raise ValueError("multivariate normal only works with 2D samples")

    return _resample_parametric(sample_np, size, dist, rng)


def bootstrap(
    fn: Callable[..., np.ndarray],
    sample: "ArrayLike",
    *args: "ArrayLike",
    **kwargs: Any,
) -> np.ndarray:
    """
    Calculate function values from bootstrap samples.

    This is equivalent to ``numpy.array([fn(b) for b in resample(sample)])`` and
    implemented for convenience.

    Parameters
    ----------
    fn : Callable
        Bootstrap samples are passed to this function.
    sample : array-like
        Original sample.
    *args : array-like
        Optional additional arrays of the same length to resample.
    **kwargs
        Keywords are forwarded to :func:`resample`.

    Returns
    -------
    np.array
        Results of `fn` applied to each bootstrap sample.

    Examples
    --------
    >>> from resample.bootstrap import bootstrap
    >>> import numpy as np
    >>> x = np.arange(10)
    >>> fx = np.mean(x)
    >>> fb = bootstrap(np.mean, x, size=10000, random_state=1)
    >>> print(f"f(x) = {fx:.1f} +/- {np.std(fb):.1f}")
    f(x) = 4.5 +/- 0.9

    """
    gen = resample(sample, *args, **kwargs)
    if args:
        return np.array([fn(*b) for b in gen])
    return np.array([fn(x) for x in gen])


def variance(
    fn: Callable[..., np.ndarray],
    sample: "ArrayLike",
    *args: "ArrayLike",
    **kwargs: Any,
) -> np.ndarray:
    """
    Calculate bootstrap estimate of variance.

    If the function returns a vector, the variance is computed elementwise.

    Parameters
    ----------
    fn : callable
        Estimator. Can be any mapping ℝⁿ → ℝᵏ, where n is the sample size
        and k is the length of the output array.
    sample : array-like
        Original sample.
    *args : array-like
        Optional additional arrays of the same length to resample.
    **kwargs
        Keyword arguments forwarded to :func:`resample`.

    Returns
    -------
    ndarray
        Bootstrap estimate of variance.

    Examples
    --------
    Compute variance of arithmetic mean.

    >>> from resample.bootstrap import variance
    >>> import numpy as np
    >>> x = np.arange(10)
    >>> v = variance(np.mean, x, size=10000, random_state=1)
    >>> f"{v:.1f}"
    '0.8'

    """
    thetas = bootstrap(fn, sample, *args, **kwargs)
    return np.var(thetas, ddof=1, axis=0)


def covariance(
    fn: Callable[..., np.ndarray],
    sample: "ArrayLike",
    *args: "ArrayLike",
    **kwargs: Any,
) -> np.ndarray:
    """
    Calculate bootstrap estimate of covariance.

    Parameters
    ----------
    fn : callable
        Estimator. Can be any mapping ℝⁿ → ℝᵏ, where n is the sample size
        and k is the length of the output array.
    sample : array-like
        Original sample.
    *args : array-like
        Optional additional arrays of the same length to resample.
    **kwargs
        Keyword arguments forwarded to :func:`resample`.

    Returns
    -------
    ndarray
        Bootstrap estimate of covariance. In general, this is a matrix, but if the
        function maps to a scalar, it is scalar as well.

    Examples
    --------
    Compute covariance of sample mean and sample variance.

    >>> from resample.bootstrap import variance
    >>> import numpy as np
    >>> x = np.arange(10)
    >>> def fn(x):
    ...     return np.mean(x), np.var(x)
    >>> np.round(covariance(fn, x, size=10000, random_state=1), 1)
    array([[0.8, 0. ],
           [0. , 5.5]])

    """
    thetas = bootstrap(fn, sample, *args, **kwargs)
    return np.cov(thetas, rowvar=False, ddof=1)


def confidence_interval(
    fn: Callable[..., np.ndarray],
    sample: "ArrayLike",
    *args: "ArrayLike",
    cl: float = 0.95,
    ci_method: str = "bca",
    **kwargs: Any,
) -> Tuple[float, float]:
    """
    Calculate bootstrap confidence intervals.

    Parameters
    ----------
    fn : callable
        Function to be bootstrapped.
    sample : array-like
        Original sample.
    *args : array-like
        Optional additional arrays of the same length to resample.
    cl : float, default : 0.95
        Confidence level. Asymptotically, this is the probability that the interval
        contains the true value.
    ci_method : str, {'bca', 'percentile'}, optional
        Confidence interval method. Default is 'bca'. See notes for details.
    **kwargs
        Keyword arguments forwarded to :func:`resample`.

    Returns
    -------
    (float, float)
        Upper and lower confidence limits.

    Examples
    --------
    Compute confidence interval for arithmetic mean.

    >>> from resample.bootstrap import confidence_interval
    >>> import numpy as np
    >>> x = np.arange(10)
    >>> a, b = confidence_interval(np.mean, x, size=10000, random_state=1)
    >>> f"{a:.1f} to {b:.1f}"
    '2.6 to 6.2'

    Notes
    -----
    Both the 'percentile' and 'bca' methods produce intervals that are invariant to
    monotonic transformations of the data values, a desirable and consistent property.

    The 'percentile' method is straightforward and useful as a fallback. The 'bca'
    method is 2nd order accurate (to O(1/n) where n is the sample size) and generally
    preferred. It computes a jackknife estimate in addition to the bootstrap, which
    increases the number of function evaluations in a direct comparison to
    'percentile'. However the increase in accuracy should compensate for this, with the
    result that less bootstrap replicas are needed overall to achieve the same accuracy.

    """
    if args and not isinstance(args[0], Collection):
        import warnings

        warnings.warn(
            "Calling confidence_interval with positional instead of keyword "
            "arguments is deprecated",
            FutureWarning,
        )

        if len(args) == 1:
            (cl,) = args
        elif len(args) == 2:
            cl, ci_method = args
        else:
            raise ValueError("too many arguments")
        args = ()

    if not 0 < cl < 1:
        raise ValueError("cl must be between zero and one")

    thetas = bootstrap(fn, sample, *args, **kwargs)
    alpha = 1 - cl

    if ci_method == "percentile":
        return _confidence_interval_percentile(thetas, alpha / 2)

    if ci_method == "bca":
        theta = fn(sample, *args)
        j_thetas = jackknife(fn, sample, *args)
        return _confidence_interval_bca(theta, thetas, j_thetas, alpha / 2)

    raise ValueError(
        f"ci_method must be 'percentile' or 'bca', but '{ci_method}' was supplied"
    )


def _resample_stratified(
    sample: np.ndarray,
    size: int,
    method: str,
    strata: np.ndarray,
    rng: np.random.Generator,
) -> Generator[np.ndarray, None, None]:
    # call resample on sub-samples and merge the replicates
    sub_samples = [sample[strata == x] for x in np.unique(strata)]
    for sub_replicates in zip(
        *[resample(s, size=size, method=method, random_state=rng) for s in sub_samples]
    ):
        yield np.concatenate(sub_replicates, axis=0)


def _resample_ordinary_1(
    sample: np.ndarray, size: int, rng: np.random.Generator
) -> Generator[np.ndarray, None, None]:
    # i.i.d. sampling from empirical cumulative distribution of sample
    n = len(sample)
    for _ in range(size):
        yield rng.choice(sample, size=n, replace=True)


def _resample_ordinary_n(
    samples: List[np.ndarray], size: int, rng: np.random.Generator
) -> Generator[np.ndarray, None, None]:
    n = len(samples[0])
    indices = np.arange(n)
    for _ in range(size):
        m = rng.choice(indices, size=n, replace=True)
        yield tuple(s[m] for s in samples)


def _resample_balanced_1(
    sample: np.ndarray, size: int, rng: np.random.Generator
) -> Generator[np.ndarray, None, None]:
    # effectively computes a random permutation of `size` concatenated
    # copies of `sample` and returns `size` equal chunks of that
    n = len(sample)
    indices = rng.permutation(n * size)
    for i in range(size):
        m = indices[i * n : (i + 1) * n] % n
        yield sample[m]


def _resample_balanced_n(
    samples: List[np.ndarray], size: int, rng: np.random.Generator
) -> Generator[np.ndarray, None, None]:
    n = len(samples[0])
    indices = rng.permutation(n * size)
    for i in range(size):
        m = indices[i * n : (i + 1) * n] % n
        yield tuple(s[m] for s in samples)


def _resample_extended_1(
    sample: np.ndarray, size: int, rng: np.random.Generator
) -> Generator[np.ndarray, None, None]:
    # randomly generates the sample size from a Poisson distribution
    n = len(sample)
    for i in range(size):
        k = rng.poisson(1, size=n)
        yield np.repeat(sample, k, axis=0)


def _resample_extended_n(
    samples: List[np.ndarray], size: int, rng: np.random.Generator
) -> Generator[np.ndarray, None, None]:
    n = len(samples[0])
    for i in range(size):
        k = rng.poisson(1, size=n)
        yield tuple(np.repeat(s, k, axis=0) for s in samples)


def _fit_parametric_family(
    dist: stats.rv_continuous, sample: np.ndarray
) -> Tuple[float, ...]:
    if dist == stats.multivariate_normal:
        # has no fit method...
        return np.mean(sample, axis=0), np.cov(sample.T, ddof=1)

    if dist in {stats.f, stats.beta}:
        fit_kwargs = {"floc": 0, "fscale": 1}
    elif dist in {stats.gamma, stats.lognorm, stats.invgauss, stats.pareto}:
        fit_kwargs = {"floc": 0}
    else:
        fit_kwargs = {}

    return dist.fit(sample, **fit_kwargs)  # type: ignore


def _resample_parametric(
    sample: np.ndarray, size: int, dist: stats.rv_continuous, rng: np.random.Generator
) -> Generator[np.ndarray, None, None]:
    n = len(sample)

    # fit parameters by maximum likelihood and sample from that
    if dist == stats.poisson:
        # - poisson has no fit method and there is no scale parameter
        # - random number generation for poisson distribution in scipy seems to be buggy
        mu = np.mean(sample)
        for _ in range(size):
            yield rng.poisson(mu, size=n)
    else:
        args = _fit_parametric_family(dist, sample)
        dist = dist(*args)
        for _ in range(size):
            yield dist.rvs(size=n, random_state=rng)


def _confidence_interval_percentile(
    thetas: np.ndarray, alpha_half: float
) -> Tuple[float, float]:
    quant = quantile_function_gen(thetas)
    return quant(alpha_half), quant(1 - alpha_half)


def _confidence_interval_bca(
    theta: float, thetas: np.ndarray, j_thetas: np.ndarray, alpha_half: float
) -> Tuple[float, float]:
    norm = stats.norm

    # bias correction; implementation notes:
    # - if prop_less is zero, z_naught would become -inf;
    #   we set z_naught to zero then (no bias)
    prop_less = np.mean(thetas < theta)  # proportion of replicates less than obs
    z_naught = norm.ppf(prop_less) if prop_less > 0 else 0.0

    # acceleration; implementation notes:
    # - np.mean returns float even if j_thetas are int,
    #   must convert type explicity to make -= operator work
    # - it is possible that all j_thetas are zero, it then follows
    #   that den and num are zero; we set acc to zero then (no acceleration)
    j_mean = np.mean(j_thetas)
    j_thetas = j_thetas.astype(j_mean.dtype, copy=False)
    j_thetas -= j_mean
    num = np.sum((-j_thetas) ** 3)
    den = np.sum(j_thetas**2)
    acc = num / (6 * den**1.5) if den > 0 else 0.0

    z_low = z_naught + norm.ppf(alpha_half)
    z_high = z_naught + norm.ppf(1 - alpha_half)

    p_low = norm.cdf(z_naught + z_low / (1 - acc * z_low))
    p_high = norm.cdf(z_naught + z_high / (1 - acc * z_high))

    quant = quantile_function_gen(thetas)
    return quant(p_low), quant(p_high)


def __getattr__(key: str) -> Any:
    for match in ("bias", "bias_corrected"):
        if key == match:
            msg = (
                f"resample.bootstrap.{match} has been removed. The implementation was "
                "discovered to be faulty, and a generic fix is not in sight. "
                "Please use resample.jackknife.bias instead."
            )
            raise NotImplementedError(msg)
    raise AttributeError



# ruff: noqa: D100 D103
import numpy as np
import pytest
from numpy.testing import assert_equal

from resample.empirical import cdf_gen, influence, quantile_function_gen


# high-quality platform-independent reproducible sequence of pseudo-random numbers
@pytest.fixture
def rng():
    return np.random.Generator(np.random.PCG64(1))


def test_cdf_increasing(rng):
    x = rng.normal(size=100)
    cdf = cdf_gen(x)
    result = [cdf(s) for s in np.linspace(x.min(), x.max(), 100)]
    assert np.all(np.diff(result) >= 0)


def test_cdf_at_infinity():
    cdf = cdf_gen(np.arange(10))
    assert cdf(-np.inf) == 0.0
    assert cdf(np.inf) == 1.0


def test_cdf_simple_cases():
    cdf = cdf_gen([0, 1, 2, 3])
    assert cdf(0) == 0.25
    assert cdf(1) == 0.5
    assert cdf(2) == 0.75
    assert cdf(3) == 1.0


def test_cdf_on_array():
    x = np.arange(4)
    cdf = cdf_gen(x)
    assert_equal(cdf(x), (x + 1) / len(x))
    assert_equal(cdf(x + 1e-10), (x + 1) / len(x))
    assert_equal(cdf(x - 1e-10), x / len(x))


def test_quantile_simple_cases():
    q = quantile_function_gen([0, 1, 2, 3])
    assert q(0.25) == 0
    assert q(0.5) == 1
    assert q(0.75) == 2
    assert q(1.0) == 3


def test_quantile_on_array():
    x = np.arange(4)
    q = quantile_function_gen(x)
    prob = (x + 1) / len(x)
    assert_equal(q(prob), x)


def test_quantile_is_inverse_of_cdf(rng):
    x = rng.normal(size=30)
    y = cdf_gen(x)(x)
    assert_equal(quantile_function_gen(x)(y), x)


@pytest.mark.parametrize("arg", [-1, 1.5])
def test_quantile_out_of_bounds_is_nan(arg):
    q = quantile_function_gen(np.array([0, 1, 2, 3]))
    assert np.isnan(q(arg))


def test_influence_shape():
    n = 100
    data = np.random.random(n)
    emp = influence(np.mean, data)
    assert len(emp) == n



"""
Empirical functions.

Empirical functions based on a data sample instead of a parameteric density function,
like the empirical CDF. Implemented here are mostly tools used internally.
"""

__all__ = ["cdf_gen", "quantile_function_gen", "influence"]

from typing import Callable, Union

import numpy as np
from numpy.typing import ArrayLike

from .jackknife import jackknife


def cdf_gen(sample: "ArrayLike") -> Callable[[np.ndarray], np.ndarray]:
    """
    Return the empirical distribution function for the given sample.

    Parameters
    ----------
    sample : array-like
        Sample.

    Returns
    -------
    callable
        Empirical distribution function.

    """
    sample_np = np.sort(sample)
    n = len(sample_np)
    return lambda x: np.searchsorted(sample_np, x, side="right", sorter=None) / n


def quantile_function_gen(
    sample: "ArrayLike",
) -> Callable[[Union[float, "ArrayLike"]], Union[float, np.ndarray]]:
    """
    Return the empirical quantile function for the given sample.

    Parameters
    ----------
    sample : array-like
        Sample.

    Returns
    -------
    callable
        Empirical quantile function.

    """

    class QuantileFn:
        def __init__(self, sample: "ArrayLike"):
            self._sorted = np.sort(sample, axis=0)

        def __call__(self, p: Union[float, "ArrayLike"]) -> Union[float, np.ndarray]:
            ndim = np.ndim(p)  # must come before atleast_1d
            p = np.atleast_1d(p)
            result = np.empty(len(p))
            valid = (p >= 0) & (p <= 1)
            n = len(self._sorted)
            idx = np.maximum(np.ceil(p[valid] * n).astype(int) - 1, 0)
            result[valid] = self._sorted[idx]
            result[~valid] = np.nan
            if ndim == 0:
                return result[0]
            return result

    return QuantileFn(sample)


def influence(
    fn: Callable[["ArrayLike"], np.ndarray], sample: "ArrayLike"
) -> np.ndarray:
    """
    Calculate the empirical influence function for a given sample and estimator.

    Parameters
    ----------
    fn : callable
        Estimator. Can be any mapping ℝⁿ → ℝᵏ, where n is the sample size
        and k is the length of the output array.
    sample : array-like
        Sample. Must be one-dimensional.

    Returns
    -------
    ndarray
        Empirical influence values.

    """
    sample = np.atleast_1d(sample)
    n = len(sample)
    return (n - 1) * (fn(sample) - jackknife(fn, sample))



"""
Resampling tools for Python.

This library offers randomisation-based inference in Python based on data resampling
and permutation. The following functionality is implemented.

- Bootstrap samples (ordinary or balanced with optional stratification) from N-D arrays
- Apply parametric bootstrap (Gaussian, Poisson, gamma, etc.) on samples
- Compute bootstrap confidence intervals (percentile or BCa) for any estimator
- Jackknife estimates of bias and variance of any estimator
- Permutation-based variants of traditional statistical tests (t-test, K-S test, etc.)
- Tools for working with empirical distributions (CDF, quantile, etc.)
"""

from importlib.metadata import version

from resample import bootstrap, empirical, jackknife, permutation

__version__ = version("resample")

__all__ = [
    "jackknife",
    "bootstrap",
    "permutation",
    "empirical",
    "__version__",
]



# ruff: noqa: D100 D103
from resample import _util as u
import numpy as np
from numpy.testing import assert_allclose


def test_wilson_score_interval():
    n = 100
    for n1 in (10, 50, 90):
        p, lh = u.wilson_score_interval(n1, n, 1)
        s = np.sqrt(p * (1 - p) / n)
        assert_allclose(p, n1 / n)
        assert_allclose(lh, (p - s, p + s), atol=0.01)

    n = 10
    n1 = 0
    p, lh = u.wilson_score_interval(n1, n, 1)
    assert_allclose(p, 0.0)
    assert_allclose(lh, (0, 0.1), atol=0.01)

    n1 = 10
    p, lh = u.wilson_score_interval(n1, n, 1)
    assert_allclose(p, 1.0)
    assert_allclose(lh, (0.9, 1.0), atol=0.01)



# -*- coding: utf-8 -*-
#
# Configuration file for the Sphinx documentation builder.
#
# This file does only contain a selection of the most common options. For a
# full list see the documentation:
# http://www.sphinx-doc.org/en/master/config

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
# import os
# import sys
# sys.path.insert(0, os.path.abspath('.'))

# -- Project information -----------------------------------------------------
import resample

version = resample.__version__  # noqa

project = "resample"
copyright = "2018, Daniel Saxton"
author = "Daniel Saxton and Hans Dembinski"

# -- General configuration ---------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#
# needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.napoleon",
    "nbsphinx",
]

autoclass_content = "both"
autosummary_generate = True
autodoc_member_order = "groupwise"
autodoc_type_aliases = {"ArrayLike": "ArrayLike"}

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
# source_suffix = ['.rst', '.md']
source_suffix = ".rst"

# The master toctree document.
master_doc = "index"

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = "en"

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = [
    "_build",
    "Thumbs.db",
    ".DS_Store",
    "__pycache__",
    ".ipynb_checkpoints",
]

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = None


# -- Options for HTML output -------------------------------------------------

import sphinx_rtd_theme

html_theme = "sphinx_rtd_theme"
html_theme_path = [sphinx_rtd_theme.get_html_theme_path()]

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.

html_logo = "_static/logo.svg"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]

# Custom sidebar templates, must be a dictionary that maps document names
# to template names.
#
# The default sidebars (for documents that don't match any pattern) are
# defined by theme itself.  Builtin themes are using these templates by
# default: ``['localtoc.html', 'relations.html', 'sourcelink.html',
# 'searchbox.html']``.
#
# html_sidebars = {}

# -- Options for HTMLHelp output ---------------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = "resampledoc"



import numpy as np
import pytest

from scipy.stats import random_table


@pytest.mark.parametrize("n", (10, 100, 1000, 10000, 100000))
@pytest.mark.parametrize("k", (2, 4, 10, 20, 40, 100))
@pytest.mark.parametrize("method", (None, "boyett", "patefield"))
def test_rcont(k, n, method, benchmark):
    w = np.zeros((k, k))
    rng = np.random.default_rng(1)
    for _ in range(n):
        i = rng.integers(k)
        j = rng.integers(k)
        w[i, j] += 1
    r = np.sum(w, axis=1)
    c = np.sum(w, axis=0)
    assert np.sum(r) == n
    assert np.sum(c) == n
    benchmark(lambda: random_table(r, c).rvs(100, method=method, random_state=rng))



import numpy as np
import pytest
from numpy.testing import assert_almost_equal, assert_equal
from scipy.optimize import curve_fit
from resample.jackknife import (
    bias,
    bias_corrected,
    jackknife,
    resample,
    variance,
    cross_validation,
)


def test_resample_1d():
    data = (1, 2, 3)

    r = []
    for x in resample(data):
        r.append(x.copy())
    assert_equal(r, [[2, 3], [1, 3], [1, 2]])


def test_resample_2d():
    data = ((1, 2), (3, 4), (5, 6))

    r = []
    for x in resample(data):
        r.append(x.copy())
    assert_equal(r, [[(3, 4), (5, 6)], [(1, 2), (5, 6)], [(1, 2), (3, 4)]])


def test_jackknife():
    data = (1, 2, 3)
    r = jackknife(lambda x: x.copy(), data)
    assert_equal(r, [[2, 3], [1, 3], [1, 2]])


def test_bias_on_unbiased():
    data = (0, 1, 2, 3)
    # bias is exactly zero for linear functions
    r = bias(np.mean, data)
    assert r == 0


def test_bias_on_biased_order_n_minus_one():
    # this "mean" has a bias of exactly O(n^{-1})
    def bad_mean(x):
        return (np.sum(x) + 2) / len(x)

    data = (0, 1, 2)
    r = bias(bad_mean, data)
    mean_jk = np.mean([bad_mean([1, 2]), bad_mean([0, 2]), bad_mean([0, 1])])
    # (5/2 + 4/2 + 3/2) / 3 = 12 / 6 = 2
    assert mean_jk == 2.0
    # f = 5/3
    # (n-1) * (mean_jk - f)
    # (3 - 1) * (6/3 - 5/3) = 2/3
    # note: 2/3 is exactly the bias of bad_mean for n = 3
    assert r == pytest.approx(2.0 / 3.0)


def test_bias_on_array_map():
    # compute mean and (biased) variance simultanously
    def fn(x):
        return np.mean(x), np.var(x, ddof=0)

    data = (0, 1, 2)
    r = bias(fn, data)
    assert_almost_equal(r, (0.0, -1.0 / 3.0))


def test_bias_corrected():
    # this "mean" has a bias of exactly O(n^{-1})
    def bad_mean(x):
        return (np.sum(x) + 2) / len(x)

    # bias correction is exact up to O(n^{-1})
    data = (0, 1, 2)
    r = bias_corrected(bad_mean, data)
    assert r == 1.0  # which is the correct unbiased mean


def test_variance():
    data = (0, 1, 2)
    r = variance(np.mean, data)
    # formula is (n - 1) / n * sum((jf - mean(jf)) ** 2)
    # fj = [3/2, 1, 1/2]
    # mfj = 1
    # ((3/2 - 1)^2 + (1 - 1)^2 + (1/2 - 1)^2) * 2 / 3
    # (1/4 + 1/4) / 3 * 2 = 1/3
    assert r == pytest.approx(1.0 / 3.0)


def test_resample_several_args():
    a = [1, 2, 3]
    b = [(1, 2), (2, 3), (3, 4)]
    c = ["12", "3", "4"]
    for ai, bi, ci in resample(a, b, c):
        assert np.shape(ai) == (2,)
        assert np.shape(bi) == (2, 2)
        assert np.shape(ci) == (2,)
        assert set(ai) <= set(a)
        assert set(ci) <= set(c)
        bi = list(tuple(x) for x in bi)
        assert set(bi) <= set(b)


def test_resample_several_args_incompatible_keywords():
    a = [1, 2, 3]
    with pytest.raises(ValueError):
        resample(a, [1, 2])

    with pytest.raises(ValueError):
        resample(a, [1, 2, 3, 4])


def test_resample_deprecation():
    data = [1, 2, 3]

    with pytest.warns(FutureWarning):
        r = list(resample(data, False))

    assert_equal(r, list(resample(data, copy=False)))

    with pytest.warns(FutureWarning):
        with pytest.raises(ValueError):  # too many arguments
            resample(data, True, 1)


@pytest.mark.filterwarnings("ignore:Covariance")
def test_cross_validation():
    x = [1, 2, 3]
    y = [3, 4, 5]

    def predict(xi, yi, xo, npar):
        def model(x, *par):
            return np.polyval(par, x)

        popt = curve_fit(model, xi, yi, p0=np.zeros(npar))[0]
        return model(xo, *popt)

    v = cross_validation(predict, x, y, 2)
    assert v == pytest.approx(0)

    v2 = cross_validation(predict, x, y, 1)
    assert v2 == pytest.approx(1.5)



import numpy as np
import pytest

from resample.bootstrap import confidence_interval, resample


def run_resample(n, method):
    x = np.arange(n)
    r = []
    for b in resample(x, method=method):
        r.append(b)
    return r


@pytest.mark.benchmark(group="bootstrap-100")
@pytest.mark.parametrize("method", ("ordinary", "balanced", "normal"))
def test_resample_100(benchmark, method):
    benchmark(run_resample, 100, method)


@pytest.mark.benchmark(group="bootstrap-1000")
@pytest.mark.parametrize("method", ("ordinary", "balanced", "normal"))
def test_bootstrap_resample_1000(benchmark, method):
    benchmark(run_resample, 1000, method)


@pytest.mark.benchmark(group="bootstrap-10000")
@pytest.mark.parametrize("method", ("ordinary", "balanced", "normal"))
def test_bootstrap_resample_10000(benchmark, method):
    benchmark(run_resample, 10000, method)


def run_confidence_interval(n, ci_method):
    x = np.arange(n)
    confidence_interval(np.mean, x, ci_method=ci_method)


@pytest.mark.benchmark(group="confidence-interval-100")
@pytest.mark.parametrize("ci_method", ("percentile", "bca"))
def test_bootstrap_confidence_interval_100(benchmark, ci_method):
    benchmark(run_confidence_interval, 100, ci_method)


@pytest.mark.benchmark(group="confidence-interval-1000")
@pytest.mark.parametrize("ci_method", ("percentile", "bca"))
def test_bootstrap_confidence_interval_1000(benchmark, ci_method):
    benchmark(run_confidence_interval, 1000, ci_method)


@pytest.mark.benchmark(group="confidence-interval-10000")
@pytest.mark.parametrize("ci_method", ("percentile", "bca"))
def test_bootstrap_confidence_interval_10000(benchmark, ci_method):
    benchmark(run_confidence_interval, 10000, ci_method)



# ruff: noqa: D100 D103
import numpy as np
import pytest
from numpy.testing import assert_equal, assert_allclose
from scipy import stats

from resample.bootstrap import (
    _fit_parametric_family,
    bootstrap,
    confidence_interval,
    resample,
    variance,
    covariance,
)

PARAMETRIC_CONTINUOUS = {
    # use scipy.stats names here
    "norm",
    "t",
    "laplace",
    "logistic",
    "f",
    "beta",
    "gamma",
    "lognorm",
    "invgauss",
    "pareto",
}
PARAMETRIC_DISCRETE = {"poisson"}
PARAMETRIC = PARAMETRIC_CONTINUOUS | PARAMETRIC_DISCRETE
NON_PARAMETRIC = {"ordinary", "balanced"}
ALL_METHODS = NON_PARAMETRIC | PARAMETRIC


def chisquare(
    obs, exp=None
):  # we do not use scipy.stats.chisquare, because it is broken
    n = len(obs)
    if exp is None:
        exp = 1.0 / n
    t = np.sum(obs**2 / exp) - n
    return stats.chi2(n - 1).cdf(t)


@pytest.fixture
def rng():
    return np.random.Generator(np.random.PCG64(1))


@pytest.mark.parametrize("method", ALL_METHODS)
def test_resample_shape_1d(method):
    if method == "beta":
        x = (0.1, 0.2, 0.3)
    else:
        x = (1.0, 2.0, 3.0)
    n_rep = 5
    count = 0
    with np.errstate(invalid="ignore"):
        for bx in resample(x, size=n_rep, method=method):
            assert len(bx) == len(x)
            count += 1
    assert count == n_rep


@pytest.mark.parametrize("method", NON_PARAMETRIC | {"norm"})
def test_resample_shape_2d(method):
    x = [(1.0, 2.0), (4.0, 3.0), (6.0, 5.0)]
    n_rep = 5
    count = 0
    for bx in resample(x, size=n_rep, method=method):
        assert bx.shape == np.shape(x)
        count += 1
    assert count == n_rep


@pytest.mark.parametrize("method", NON_PARAMETRIC)
def test_resample_shape_4d(method):
    x = np.ones((2, 3, 4, 5))
    n_rep = 5
    count = 0
    for bx in resample(x, size=n_rep, method=method):
        assert bx.shape == np.shape(x)
        count += 1
    assert count == n_rep


@pytest.mark.parametrize("method", NON_PARAMETRIC | PARAMETRIC_CONTINUOUS)
def test_resample_1d_statistical_test(method, rng):
    # distribution parameters for parametric families
    args = {
        "t": (2,),
        "f": (25, 20),
        "beta": (2, 1),
        "gamma": (1.5,),
        "lognorm": (1.0,),
        "invgauss": (1,),
        "pareto": (1,),
    }.get(method, ())

    if method in NON_PARAMETRIC:
        dist = stats.norm
    else:
        dist = getattr(stats, method)

    x = dist.rvs(*args, size=1000, random_state=rng)

    # make equidistant bins in quantile space for this particular data set
    with np.errstate(invalid="ignore"):
        par = _fit_parametric_family(dist, x)
    prob = np.linspace(0, 1, 11)
    xe = dist(*par).ppf(prob)

    # - in case of parametric bootstrap, wref is exactly uniform
    # - in case of ordinary and balanced, it needs to be computed from original sample
    if method in NON_PARAMETRIC:
        wref = np.histogram(x, bins=xe)[0]
    else:
        wref = len(x) / (len(xe) - 1)

    # compute P values for replicates compared to original
    prob = []
    wsum = 0
    with np.errstate(invalid="ignore"):
        for bx in resample(x, size=100, method=method, random_state=rng):
            w = np.histogram(bx, bins=xe)[0]
            wsum += w
            pvalue = chisquare(w, wref)
            prob.append(pvalue)

    if method == "balanced":
        # balanced bootstrap exactly reproduces frequencies in original sample
        assert_equal(wref * 100, wsum)

    # check whether P value distribution is flat
    # - test has chance probability of 1 % to fail randomly
    # - if it fails due to programming error, value is typically < 1e-20
    wp = np.histogram(prob, range=(0, 1))[0]
    pvalue = chisquare(wp)
    assert pvalue > 0.01


def test_resample_1d_statistical_test_poisson(rng):
    # poisson is behaving super weird in scipy
    x = rng.poisson(1.5, size=1000)
    mu = np.mean(x)

    xe = (0, 1, 2, 3, 10)
    # somehow location 1 is needed here...
    wref = np.diff(stats.poisson(mu, 1).cdf(xe)) * len(x)

    # compute P values for replicates compared to original
    prob = []
    for bx in resample(x, size=100, method="poisson", random_state=rng):
        w = np.histogram(bx, bins=xe)[0]

        pvalue = chisquare(w, wref)
        prob.append(pvalue)

    # check whether P value distribution is flat
    # - test has chance probability of 1 % to fail randomly
    # - if it fails due to programming error, value is typically < 1e-20
    wp = np.histogram(prob, range=(0, 1))[0]
    pvalue = chisquare(wp)
    assert pvalue > 0.01


def test_resample_invalid_family_raises():
    msg = "Invalid family"
    with pytest.raises(ValueError, match=msg):
        next(resample((1, 2, 3), method="foobar"))


@pytest.mark.parametrize("method", PARAMETRIC - {"norm"})
def test_resample_2d_parametric_raises(method):
    with pytest.raises(ValueError):
        next(resample(np.ones((2, 2)), method=method))


def test_resample_3d_parametric_normal_raises():
    with pytest.raises(ValueError):
        next(resample(np.ones((2, 2, 2)), method="normal"))


def test_resample_equal_along_axis():
    data = np.reshape(np.tile([0, 1, 2], 3), (3, 3))
    for b in resample(data, size=2):
        assert_equal(data, b)


@pytest.mark.parametrize("method", NON_PARAMETRIC)
def test_resample_full_strata(method):
    data = np.arange(3)
    for b in resample(data, size=2, strata=data, method=method):
        assert_equal(data, b)


def test_resample_invalid_strata_raises():
    msg = "must have the same shape"
    with pytest.raises(ValueError, match=msg):
        next(resample((1, 2, 3), strata=np.arange(4)))


def test_bootstrap_2d_balanced(rng):
    data = ((1, 2, 3), (2, 3, 4), (3, 4, 5))

    def mean(x):
        return np.mean(x, axis=0)

    r = bootstrap(mean, data, method="balanced")

    # arithmetic mean is linear, therefore mean over all replicates in
    # balanced bootstrap is equal to mean of original sample
    assert_allclose(mean(data), mean(r))


@pytest.mark.parametrize("action", [bootstrap, variance, confidence_interval])
def test_bootstrap_several_args(action):
    x = [1, 2, 3]
    y = [4, 5, 6]
    xy = np.transpose([x, y])

    if action is confidence_interval:

        def f1(x, y):
            return np.sum(x + y)

        def f2(xy):
            return np.sum(xy)

    else:

        def f1(x, y):
            return np.sum(x), np.sum(y)

        def f2(xy):
            return np.sum(xy, axis=0)

    r1 = action(f1, x, y, size=10, random_state=1)
    r2 = action(f2, xy, size=10, random_state=1)

    assert_equal(r1, r2)


@pytest.mark.parametrize("ci_method", ["percentile", "bca"])
def test_confidence_interval(ci_method, rng):
    data = rng.normal(size=1000)
    par = stats.norm.fit(data)
    dist = stats.norm(
        par[0], par[1] / len(data) ** 0.5
    )  # accuracy of mean is sqrt(n) better
    cl = 0.9
    ci_ref = dist.ppf(0.05), dist.ppf(0.95)
    ci = confidence_interval(
        np.mean, data, cl=cl, size=1000, ci_method=ci_method, random_state=rng
    )
    assert_allclose(ci_ref, ci, atol=6e-3)


def test_confidence_interval_invalid_p_raises():
    msg = "must be between zero and one"
    with pytest.raises(ValueError, match=msg):
        confidence_interval(np.mean, (1, 2, 3), cl=2)


def test_confidence_interval_invalid_ci_method_raises():
    msg = "method must be 'percentile' or 'bca'"
    with pytest.raises(ValueError, match=msg):
        confidence_interval(np.mean, (1, 2, 3), ci_method="foobar")


def test_bca_confidence_interval_estimator_returns_int(rng):
    def fn(data):
        return int(np.mean(data))

    data = (1, 2, 3)
    ci = confidence_interval(fn, data, ci_method="bca", size=5, random_state=rng)
    assert_allclose((1.0, 2.0), ci)


@pytest.mark.parametrize("ci_method", ["percentile", "bca"])
def test_bca_confidence_interval_bounded_estimator(ci_method, rng):
    def fn(data):
        return max(np.mean(data), 0)

    data = (-3, -2, -1)
    ci = confidence_interval(fn, data, ci_method=ci_method, size=5, random_state=rng)
    assert_allclose((0.0, 0.0), ci)


@pytest.mark.parametrize("method", NON_PARAMETRIC)
def test_variance(method, rng):
    data = np.arange(100)
    v = np.var(data) / len(data)

    r = variance(np.mean, data, size=1000, method=method, random_state=rng)
    assert r == pytest.approx(v, rel=0.05)


@pytest.mark.parametrize("method", NON_PARAMETRIC)
def test_covariance(method, rng):
    cov = np.array([[1.0, 0.1], [0.1, 2.0]])
    data = rng.multivariate_normal([0.1, 0.2], cov, size=1000)

    r = covariance(
        lambda x: np.mean(x, axis=0), data, size=1000, method=method, random_state=rng
    )
    assert_allclose(r, cov / len(data), atol=1e-3)


def test_resample_deprecation(rng):
    data = [1, 2, 3]

    with pytest.warns(FutureWarning):
        r = list(resample(data, 10))
        assert np.shape(r) == (10, 3)

    with pytest.warns(FutureWarning):
        resample(data, 10, "balanced")

    with pytest.warns(FutureWarning):
        with pytest.raises(ValueError):
            resample(data, 10, "foo")

    with pytest.warns(FutureWarning):
        resample(data, 10, "balanced", [1, 1, 2])

    with pytest.warns(FutureWarning):
        with pytest.raises(ValueError):
            resample(data, 10, "balanced", [1, 1])

    with pytest.warns(FutureWarning):
        resample(data, 10, "balanced", [1, 1, 2], rng)

    with pytest.warns(FutureWarning):
        resample(data, 10, "balanced", [1, 1, 2], 1)

    with pytest.warns(FutureWarning):
        with pytest.raises(TypeError):
            resample(data, 10, "balanced", [1, 1, 2], 1.3)

    with pytest.warns(FutureWarning):
        with pytest.raises(ValueError):  # too many arguments
            resample(data, 10, "balanced", [1, 1, 2], 1, 2)


def test_confidence_interval_deprecation(rng):
    d = [1, 2, 3]
    with pytest.warns(FutureWarning):
        r = confidence_interval(np.mean, d, 0.6, random_state=1)
    assert_equal(r, confidence_interval(np.mean, d, cl=0.6, random_state=1))

    with pytest.warns(FutureWarning):
        r = confidence_interval(np.mean, d, 0.6, "percentile", random_state=1)
    assert_equal(
        r,
        confidence_interval(np.mean, d, cl=0.6, ci_method="percentile", random_state=1),
    )

    with pytest.warns(FutureWarning):
        with pytest.raises(ValueError):
            confidence_interval(np.mean, d, 0.6, "percentile", 1)


def test_random_state():
    d = [1, 2, 3]
    a = list(resample(d, size=5, random_state=np.random.default_rng(1)))
    b = list(resample(d, size=5, random_state=1))
    c = list(resample(d, size=5, random_state=[2, 3]))
    assert_equal(a, b)
    assert not np.all([np.all(ai == ci) for (ai, ci) in zip(a, c)])

    with pytest.raises(TypeError):
        resample(d, size=5, random_state=1.5)


@pytest.mark.parametrize("method", NON_PARAMETRIC)
def test_resample_several_args(method):
    a = [1, 2, 3]
    b = [(1, 2), (2, 3), (3, 4)]
    c = ["12", "3", "4"]
    r1 = [[], [], []]
    for ai, bi, ci in resample(a, b, c, size=5, method=method, random_state=1):
        r1[0].append(ai)
        r1[1].append(bi)
        r1[2].append(ci)

    r2 = [[], [], []]
    abc = np.empty(3, dtype=[("a", "i"), ("b", "i", 2), ("c", "U4")])
    abc[:]["a"] = a
    abc[:]["b"] = b
    abc[:]["c"] = c
    for abci in resample(abc, size=5, method=method, random_state=1):
        r2[0].append(abci["a"])
        r2[1].append(abci["b"])
        r2[2].append(abci["c"])

    for i in range(3):
        assert_equal(r1[i], r2[i])


def test_resample_several_args_incompatible_keywords():
    a = [1, 2, 3]
    b = [(1, 2), (2, 3), (3, 4)]
    with pytest.raises(ValueError):
        resample(a, b, size=5, method="norm")

    resample(a, size=5, strata=[1, 1, 2])

    with pytest.raises(ValueError):
        resample(a, b, size=5, strata=[1, 1, 2])

    resample(a, b, a, b, size=5)

    with pytest.raises(ValueError):
        resample(a, [1, 2])

    with pytest.raises(ValueError):
        resample(a, [1, 2, 3, 4])

    with pytest.raises(ValueError):
        resample(a, b, 5)


def test_resample_extended_1():
    a = [1, 2, 3]
    bs = list(resample(a, size=100, method="extended", random_state=1))

    # check that lengths of bootstrap samples are poisson distributed
    w, xe = np.histogram([len(b) for b in bs], bins=10, range=(0, 10))
    wm = stats.poisson(len(a)).pmf(xe[:-1]) * np.sum(w)
    t = np.sum((w - wm) ** 2 / wm)
    pvalue = 1 - stats.chi2(len(w)).cdf(t)
    assert pvalue > 0.1


def test_resample_extended_2():
    n = 10
    a = np.arange(n)
    ts = []
    for b in resample(a, size=1000, method="extended", random_state=1):
        ts.append(np.mean(b))

    t = np.var(ts)
    expected_not_extended = np.var(a) / n

    k = np.arange(100)
    pk = stats.poisson(n).pmf(k)
    expected = expected_not_extended * np.sum(pk[1:] * n / k[1:]) / (1 - pk[0])

    assert expected / expected_not_extended > 1.1
    assert t > expected_not_extended
    assert_allclose(t, expected, atol=0.02)


def test_resample_extended_3():
    n = 10
    a = np.arange(n)
    b = 5 + a
    ns = []
    for ai, bi in resample(a, b, size=1000, method="extended", random_state=1):
        assert len(ai) == len(bi)
        assert_equal(bi - ai, 5)
        ns.append(len(ai))
    assert_allclose(np.var(ns), 10, rtol=0.05)


def test_resample_extended_4():
    x = np.ones(10)
    a = np.transpose((x, 3 * x))

    ts = []
    for b in resample(a, size=1000, method="extended", random_state=1):
        ts.append(np.sum(b, axis=0))

    t = np.var(ts, axis=0)

    mu = np.sum(x, axis=0)
    assert_allclose(t, (mu, 3**2 * mu), rtol=0.05)


def test_resample_extended_5():
    x = np.ones(10)
    a = np.transpose((x, 3 * x))

    ts1 = []
    ts2 = []
    for b1, b2 in resample(a, 3 * a, size=1000, method="extended", random_state=1):
        ts1.append(np.sum(b1, axis=0))
        ts2.append(np.sum(b2, axis=0))

    t1 = np.var(ts1, axis=0)
    t2 = np.var(ts2, axis=0)

    mu1 = np.sum(x, axis=0)
    mu2 = 3**2 * np.sum(x, axis=0)
    assert_allclose(t1, (mu1, 3**2 * mu1), rtol=0.05)
    assert_allclose(t2, (mu2, 3**2 * mu2), rtol=0.05)


def test_bias_error():
    with pytest.raises(NotImplementedError):
        from resample.bootstrap import bias  # noqa

    with pytest.raises(NotImplementedError):
        import resample.bootstrap as b

        b.bias_corrected  # noqa



import numpy as np
import pytest

from resample.permutation import usp


@pytest.mark.parametrize("n", (10, 100, 1000, 10000))
@pytest.mark.parametrize("k", (2, 10, 100))
@pytest.mark.parametrize("method", ("patefield", "shuffle"))
def test_usp(k, n, method, benchmark):
    w = np.zeros((k, k))
    rng = np.random.default_rng(1)
    for _ in range(n):
        i = rng.integers(k)
        j = rng.integers(k)
        w[i, j] += 1
    assert np.sum(w) == n
    benchmark(lambda: usp(w, method=method, size=100, random_state=1))



"""
Generate prototypes of the resample logo.

The final logo was further edited in Inkscape. The chosen version uses the Gentium Plus
Regular font from https://fontlibrary.org/en/font/gentium-plus#Gentium%20Plus-Regular.
"""

import numpy as np
from matplotlib import pyplot as plt

for font, x0 in (("ubuntu", 0.2), ("gentium", 0.17)):

    plt.figure(figsize=(5, 1.4))
    ax = plt.subplot()
    for k in ax.spines:
        ax.spines[k].set_visible(False)
    plt.tick_params(
        **{k: False for k in ax.spines}, **{f"label{k}": False for k in ax.spines}
    )
    plt.gca().set_facecolor("none")

    size = 70
    w = 0.05
    h = 0.15
    y0 = 0.1

    # original
    plt.figtext(0, y0, "re", color="r", name=font, size=size, weight="bold")
    plt.figtext(x0, y0, "sample", color="0.2", name=font, size=size)

    # copies
    rng = np.random.default_rng(1)
    s = np.fromiter("resample", "U1")
    n = 2
    for i, col in enumerate(("0.8", "0.9")):
        x = (i + 1) * w
        y = y0 + (i + 1) * h
        s2 = rng.choice(s, size=len(s))
        plt.figtext(x, y, "".join(s2), color=col, name=font, size=size, zorder=-(i + 1))

    plt.savefig(f"{font}.svg")
